# ShanghaiTech CS110 2019 Project 1.2

#==============================================================================
#                              Project 1-2 Part 2
#                               SymbolList README
#==============================================================================
# In this file you will implement a linked list-based data structure for
# storing symbols. Its purpose is analogous to SymbolTable from Proj1-1, but to
# distinguish it we will call it SymbolList.
#
# Each node in the SymbolList contains a (addr, name) pair. An empty SymbolList
# is simply a pointer to NULL. As (addr, name) pairs are added to a SymbolList,
# new list nodes are dynamically allocated and appended to the front of the list. 
# However, as there is no free() in Venus, you do not need to write a free_list()
# function. You may use the functions in string.s during your implementation. 
# What you should do is paste specified function into corresponding position.
#
# You do not need to perform error checking on duplicate addr or name entries
# in SymbolList, nor do you need to do alignment checking. Don't worry about
# provided addresses being too big.
#
# If SymbolList were to be written in C, each list node would be declared as 
# follows:
#   typedef struct symbollist { 
#       int addr;
#       char* name;
#       struct symbollist* next; 
#   } SymbolList;
#
# You need to write new_node(), symbol_for_addr(), addr_for_symbol() and add_to_list().
# You will find test cases in this file. If you want to change test partition, 
# make sure you have totally understood the test case before you changing them.
#==============================================================================

.data
#-------------------------------------------------------
# Test Data - Feel free to add your own after this line
#-------------------------------------------------------
# Test strings
test_label1:    .asciiz "Label 1"
test_label2:    .asciiz "Label 2"
test_label3:    .asciiz "Label 3"

# Test list
node1:      .word 1234 test_label1 0
node2:      .word 3456 test_label2 node1
node3:      .word 5678 test_label3 node2
#-------------------------------------------------------
# Test Data - Feel free to add your own above this line
#-------------------------------------------------------

test_header_name:   
        .asciiz "Running symbol list tests:\n"
test_symbol_for_addr_name:
        .asciiz "Testing symbol_for_addr():\n"
test_addr_for_symbol_name:
        .asciiz "Testing addr_for_symbol():\n"
test_add_to_list_name:  
        .asciiz "Testing add_to_list():\n"

newline:   .asciiz "\n"
tab:       .asciiz "\t"

tc_actual_str:      .asciiz "actual:   "
tc_expected_str:    .asciiz "expected: "
tc_testpass_str:    .asciiz " ...test passes\n"
tc_testfail_str:    .asciiz " ...test FAILS\n"

.globl main
.text
#------------------------------------------------------------------------------
# function new_node()
#------------------------------------------------------------------------------
# Creates a new uninitialized SymbolList node on heap.
# Arguments: none
# Returns: pointer to a SymbolList node
#------------------------------------------------------------------------------
new_node:   
    ### YOUR CODE ###

#------------------------------------------------------------------------------
# function symbol_for_addr()
#------------------------------------------------------------------------------
#  Iterates through the SymbolList and searches for an entry with the given addr.
#  If an entry is found, return a pointer to the name. Otherwise return NULL.
#
#  Arguments:
#   a1 = addr to look for
#   a2 = pointer to a SymbolList (NULL indicates empty list)
#
#  Returns: a pointer to the name if found or NULL if not found
#------------------------------------------------------------------------------
symbol_for_addr:
    ### YOUR CODE ###

#------------------------------------------------------------------------------
# function addr_for_symbol()
#------------------------------------------------------------------------------
# Iterates through the SymbolList and searches for an entry with the given name.
# If an entry is found, return that addr. Otherwise return -1.
#
# Arguments:
#  a1 = pointer to a SymbolList (NULL indicates empty list)
#  a2 = name to look for
#
# Returns:  addr (int) of symbol if found or -1 if not found
#------------------------------------------------------------------------------
addr_for_symbol:
    ### YOUR CODE ###

#------------------------------------------------------------------------------
# function add_to_list()
#------------------------------------------------------------------------------
# Adds a (name, addr) pair to the FRONT of the list. You should call new_node()
# to create a new node. You do not need to perform error checking on duplicate 
# addr or name entries in SymbolList, nor do you need to do alignment checking.
#
# As with Project 1-1, you WILL need to create a copy of the string that was
# passed in. copy_of_str() from Part 1 will be useful. After new entry has been
# added to the list, return the new list pointer.
#
# Arguments:
#   a1 = ptr to list (may be NULL)
#   a2 = address of symbol (integer)
#   a3 = pointer to name of symbol (string)
#
# Returns: the new list pointer
#------------------------------------------------------------------------------
add_to_list:    
    ### YOUR CODE ###

###############################################################################
#                 DO NOT MODIFY ANYTHING BELOW THIS POINT                       
###############################################################################
#------------------------------------------------------------------------------
# function print_list() - DO NOT MODIFY THIS FUNCTION
#------------------------------------------------------------------------------
# Arguments:
#  a1 = pointer to a SymbolList (NULL indicates empty list)
#  a2 = print function
#------------------------------------------------------------------------------
print_list:
    addi sp, sp, -12        # Begin print_list
    sw s2, 8(sp)
    sw s1, 4(sp)
    sw ra, 0(sp)
    mv s1, a1
    mv s2, a2
print_list_loop:
    beq s1, zero, print_list_end
    lw a1, 0(s1)
    lw a2, 4(s1)
    jalr s2
    lw s1, 8(s1)
    j print_list_loop
print_list_end:
    lw s2, 8(sp)
    lw s1, 4(sp)
    lw ra, 0(sp)
    addi sp, sp, 12
    jr ra           # End print_list    

#------------------------------------------------------------------------------
# function print_symbol() - DO NOT MODIFY THIS FUNCTION
#------------------------------------------------------------------------------
# Prints one symbol to standard output.
#
# Arguments:
#  a1 = addr of symbol
#  a2 = name of symbol
#
# Returns: none
#------------------------------------------------------------------------------
print_symbol:
    li a0, 1           # Begin print_symbol()
    ecall
    la a1, tab
    li a0, 4
    ecall
    mv a1, a2
    ecall
    la a1, newline
    ecall
    jr ra           # End print_symbol()

###############################################################################
#                 DO NOT MODIFY ANYTHING ABOVE THIS POINT                       
###############################################################################


#------------------------------------------------------------------------------
# function strlen() - paste your implementation here
#------------------------------------------------------------------------------
# Arguments:
#  a1 = string input
#
# Returns: the length of the string
#------------------------------------------------------------------------------
strlen:
    ### PASTE YOUR CODE IN string.s HERE ###

#------------------------------------------------------------------------------
# function strncpy() - paste your implementation here
#------------------------------------------------------------------------------
# Arguments:
#  a1 = pointer to destination array
#  a2 = source string
#  a3 = number of characters to copy
#
# Returns: the destination array
#------------------------------------------------------------------------------
strncpy:
    ### PASTE YOUR CODE IN string.s HERE ###

#------------------------------------------------------------------------------
# function copy_of_str() - paste your implementation here
#------------------------------------------------------------------------------
# Creates a copy of a string. You will need to use sbrk (ecall 9) to allocate
# space for the string. strlen() and strncpy() will be helpful for this function.
# In Venus, to malloc memory use the sbrk ecall (ecall 9). See wiki for details.
#
# Arguments:
#   a1 = string to copy
#
# Returns: pointer to the copy of the string
#------------------------------------------------------------------------------
copy_of_str:
    ### PASTE YOUR CODE IN string.s HERE ###

#------------------------------------------------------------------------------
# function streq() - paste your implementation here
#------------------------------------------------------------------------------
# Arguments:
#  a1 = string 1
#  a2 = string 2
#
# Returns: 0 if string 1 and string 2 are equal, -1 if they are not equal
#------------------------------------------------------------------------------
streq:
    ### PASTE YOUR CODE IN string.s HERE ###

###############################################################################
#                 Sample test cases, BE CAREFUL to change them                       
###############################################################################    
#-------------------------------------------
# Prints a newline 
# CLOBBERS: a1 and a0
#-------------------------------------------    
print_newline:
    la a1, newline
    li a0, 4
    ecall
    jr ra

#-------------------------------------------
# Prints the string
#  a1 = label containing address of string
# CLOBBERS: a1 and a0
#-------------------------------------------
print_str:
    li a0, 4
    ecall
    jr ra
    
#-------------------------------------------
# Checks whether the actual result matches the expected result
# Treats arguments as signed ints
#  a1 = register containing actual result
#  a2 = expected value of register
# CLOBBERS: a2, a1, a0, and t0
#-------------------------------------------
check_int_equals:
    addi sp, sp, -4
    sw ra, 0(sp)

    mv t0, a1
    la a1, tc_actual_str
    jal print_str
    
    mv a1, t0
    li a0, 1
    ecall

    jal print_newline

    la a1, tc_expected_str
    jal print_str
    
    mv a1, a2
    li a0, 1
    ecall
    
    bne t0, a1, check_int_notequal
    jal print_newline
    la a1, tc_testpass_str
    jal print_str
    j check_int_end
check_int_notequal:
    jal print_newline
    la a1, tc_testfail_str
    jal print_str
check_int_end:    
    lw ra, 0(sp)
    addi sp, sp, 4
    jr ra

#-------------------------------------------
# Checks whether the actual result matches the expected result
# Arguments should be strings
#  a1 = register containing actual result
#  a2 = label to expected string
# CLOBBERS: a2, a1, a0, and t0
#-------------------------------------------
check_str_equals:
    addi sp, sp, -4
    sw ra, 0(sp)
    mv t0, a1

    la a1, tc_actual_str
    jal print_str

    mv a1, t0
    li a0, 4
    ecall

    jal print_newline

    la a1, tc_expected_str
    jal print_str

    mv a1, a2
    li a0, 4
    ecall

    jal print_newline

    mv a1, t0
    jal streq
    bne a0, zero, check_str_notequal
    la a1, tc_testpass_str
    jal print_str
    j check_str_end
check_str_notequal:
    la a1, tc_testfail_str
    jal print_str 
check_str_end:
    lw ra, 0(sp)
    addi sp, sp, 4
    jr ra

#-------------------------------------------
# Test driver
#-------------------------------------------
main:   
    la a1, test_header_name
    jal print_str
    jal print_newline

    la a1, node3
    jal test_symbol_for_addr
    jal print_newline

    la a1, node3
    jal test_addr_for_symbol
    jal print_newline
    
    jal test_add_to_list
    
    li a0, 10
    ecall

#-------------------------------------------
# Tests symbol_for_addr()
# a1 = address of beginning of list
#-------------------------------------------
test_symbol_for_addr:
    addi sp, sp, -8
    sw ra, 0(sp)
    sw a1, 4(sp)
    la a1, test_symbol_for_addr_name
    jal print_str

    li a1, 1234
    lw a2, 4(sp)
    jal symbol_for_addr
    mv a1, a0
    la a2, test_label1
    jal check_str_equals

    lw ra, 0(sp)
    addi sp, sp, 8
    jr ra

#-------------------------------------------
# Tests addr_for_symbol()
# a1 = address of beginning of list
#-------------------------------------------
test_addr_for_symbol:
    addi sp, sp, -12
    sw s1, 8(sp)
    sw a1, 4(sp)
    sw ra, 0(sp)
    la a1, test_addr_for_symbol_name
    jal print_str

    lw a1, 4(sp)
    la a2, test_label1
    jal addr_for_symbol
    mv a1, a0
    li a2, 1234
    jal check_int_equals

    lw ra, 0(sp)
    lw s1, 8(sp)
    addi sp, sp, 12
    jr ra

#-------------------------------------------
# Tests add_to_list()
# a1 = address of beginning of list
#
# Because the print is more complex than
# previous tests, the auto comparison haven't
# been implemented. You should check it yourself.
#-------------------------------------------
test_add_to_list:               # Begin test_add_to_list
    addi sp, sp, -4
    sw ra, 0(sp)
    
    la a1, test_add_to_list_name
    jal print_str
    li a1, 0            
    li a2, 1234
    la a3, test_label1
    jal add_to_list         # Test label 1
    mv a1, a0
    li a2, 3456
    la a3, test_label2
    jal add_to_list         # Test label 2
    mv a1, a0
    li a2, 5678
    la a3, test_label3          
    jal add_to_list         # Test label 3
    
    mv a1, a0               # print your list
    la a2, print_symbol
    jal print_list

    jal print_newline

    la a1, node3            # print expected list
    la a2, print_symbol
    jal print_list

    lw ra, 0(sp)
    addi sp, sp, 4
    jr ra
